@@ -2,6 +2,7 @@
  * Orthanc - A Lightweight, RESTful DICOM Store
  * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
  * Department, University Hospital of Liege, Belgium
+ * Copyright (C) 2017-2019 Osimis S.A., Belgium
  *
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -49,4 +50,27 @@ namespace Orthanc
     {
     }
   };
+  
+
+  /**
+   * This class is a simple implementation of a IDynamicObject that
+   * stores a single typed value.
+   */
+  template <typename T>
+  class SingleValueObject : public IDynamicObject
+  {
+  private:
+    T  value_;
+    
+  public:
+    explicit SingleValueObject(const T& value) :
+      value_(value)
+    {
+    }
+
+    const T& GetValue() const
+    {
+      return value_;
+    }
+  };
 }
