@@ -20,7 +20,7 @@
  * you do not wish to do so, delete this exception statement from your
  * version. If you delete this exception statement from all source files
  * in the program, then also delete it here.
- * 
+ *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
@@ -43,9 +43,17 @@
 #include <string.h>
 #include <limits>
 #include <stdint.h>
+#include <algorithm>
 
 namespace Orthanc
 {
+  double ImageProcessing::ImagePoint::GetDistanceTo(const ImagePoint& other) const
+  {
+    double dx = (double)(other.GetX() - GetX());
+    double dy = (double)(other.GetY() - GetY());
+    return sqrt(dx * dx + dy * dy);
+  }
+
   template <typename TargetType, typename SourceType>
   static void ConvertInternal(ImageAccessor& target,
                               const ImageAccessor& source)
@@ -147,8 +155,8 @@ namespace Orthanc
       {
         // Y = 0.2126 R + 0.7152 G + 0.0722 B
         int32_t v = (2126 * static_cast<int32_t>(s[0]) +
-                     7152 * static_cast<int32_t>(s[1]) +
-                     0722 * static_cast<int32_t>(s[2])) / 10000;
+            7152 * static_cast<int32_t>(s[1]) +
+            0722 * static_cast<int32_t>(s[2])) / 10000;
         
         if (static_cast<int32_t>(v) < static_cast<int32_t>(minValue))
         {
@@ -168,7 +176,7 @@ namespace Orthanc
 
 
   static void MemsetZeroInternal(ImageAccessor& image)
-  {      
+  {
     const unsigned int height = image.GetHeight();
     const size_t lineSize = image.GetBytesPerPixel() * image.GetWidth();
     const size_t pitch = image.GetPitch();
@@ -535,8 +543,8 @@ namespace Orthanc
         for (unsigned int x = 0; x < width; x++, q++)
         {
           *q = static_cast<uint8_t>((2126 * static_cast<uint32_t>(p[0]) +
-                                     7152 * static_cast<uint32_t>(p[1]) +
-                                     0722 * static_cast<uint32_t>(p[2])) / 10000);
+                                    7152 * static_cast<uint32_t>(p[1]) +
+              0722 * static_cast<uint32_t>(p[2])) / 10000);
           p += 4;
         }
       }
@@ -554,8 +562,8 @@ namespace Orthanc
         for (unsigned int x = 0; x < width; x++, q++)
         {
           *q = static_cast<uint8_t>((2126 * static_cast<uint32_t>(p[2]) +
-                                     7152 * static_cast<uint32_t>(p[1]) +
-                                     0722 * static_cast<uint32_t>(p[0])) / 10000);
+                                    7152 * static_cast<uint32_t>(p[1]) +
+              0722 * static_cast<uint32_t>(p[0])) / 10000);
           p += 4;
         }
       }
@@ -788,33 +796,33 @@ namespace Orthanc
   {
     switch (image.GetFormat())
     {
-      case PixelFormat_Grayscale8:
-        SetInternal<uint8_t>(image, value);
-        return;
+    case PixelFormat_Grayscale8:
+      SetInternal<uint8_t>(image, value);
+      return;
 
-      case PixelFormat_Grayscale16:
-        SetInternal<uint16_t>(image, value);
-        return;
+    case PixelFormat_Grayscale16:
+      SetInternal<uint16_t>(image, value);
+      return;
 
-      case PixelFormat_Grayscale32:
-        SetInternal<uint32_t>(image, value);
-        return;
+    case PixelFormat_Grayscale32:
+      SetInternal<uint32_t>(image, value);
+      return;
 
-      case PixelFormat_Grayscale64:
-        SetInternal<uint64_t>(image, value);
-        return;
+    case PixelFormat_Grayscale64:
+      SetInternal<uint64_t>(image, value);
+      return;
 
-      case PixelFormat_SignedGrayscale16:
-        SetInternal<int16_t>(image, value);
-        return;
+    case PixelFormat_SignedGrayscale16:
+      SetInternal<int16_t>(image, value);
+      return;
 
-      case PixelFormat_Float32:
-        assert(sizeof(float) == 4);
-        SetInternal<float>(image, value);
-        return;
+    case PixelFormat_Float32:
+      assert(sizeof(float) == 4);
+      SetInternal<float>(image, value);
+      return;
 
-      default:
-        throw OrthancException(ErrorCode_NotImplemented);
+    default:
+      throw OrthancException(ErrorCode_NotImplemented);
     }
   }
 
@@ -830,32 +838,32 @@ namespace Orthanc
 
     switch (image.GetFormat())
     {
-      case PixelFormat_RGBA32:
-        p[0] = red;
-        p[1] = green;
-        p[2] = blue;
-        p[3] = alpha;
-        size = 4;
-        break;
+    case PixelFormat_RGBA32:
+      p[0] = red;
+      p[1] = green;
+      p[2] = blue;
+      p[3] = alpha;
+      size = 4;
+      break;
 
-      case PixelFormat_BGRA32:
-        p[0] = blue;
-        p[1] = green;
-        p[2] = red;
-        p[3] = alpha;
-        size = 4;
-        break;
+    case PixelFormat_BGRA32:
+      p[0] = blue;
+      p[1] = green;
+      p[2] = red;
+      p[3] = alpha;
+      size = 4;
+      break;
 
-      case PixelFormat_RGB24:
-        p[0] = red;
-        p[1] = green;
-        p[2] = blue;
-        size = 3;
-        break;
+    case PixelFormat_RGB24:
+      p[0] = red;
+      p[1] = green;
+      p[2] = blue;
+      size = 3;
+      break;
 
-      default:
-        throw OrthancException(ErrorCode_NotImplemented);
-    }    
+    default:
+      throw OrthancException(ErrorCode_NotImplemented);
+    }
 
     const unsigned int width = image.GetWidth();
     const unsigned int height = image.GetHeight();
@@ -898,44 +906,44 @@ namespace Orthanc
   {
     switch (image.GetFormat())
     {
-      case PixelFormat_Grayscale8:
-      {
-        uint8_t a, b;
-        GetMinMaxValueInternal<uint8_t>(a, b, image);
-        minValue = a;
-        maxValue = b;
-        break;
-      }
+    case PixelFormat_Grayscale8:
+    {
+      uint8_t a, b;
+      GetMinMaxValueInternal<uint8_t>(a, b, image);
+      minValue = a;
+      maxValue = b;
+      break;
+    }
 
-      case PixelFormat_Grayscale16:
-      {
-        uint16_t a, b;
-        GetMinMaxValueInternal<uint16_t>(a, b, image);
-        minValue = a;
-        maxValue = b;
-        break;
-      }
+    case PixelFormat_Grayscale16:
+    {
+      uint16_t a, b;
+      GetMinMaxValueInternal<uint16_t>(a, b, image);
+      minValue = a;
+      maxValue = b;
+      break;
+    }
 
-      case PixelFormat_Grayscale32:
-      {
-        uint32_t a, b;
-        GetMinMaxValueInternal<uint32_t>(a, b, image);
-        minValue = a;
-        maxValue = b;
-        break;
-      }
+    case PixelFormat_Grayscale32:
+    {
+      uint32_t a, b;
+      GetMinMaxValueInternal<uint32_t>(a, b, image);
+      minValue = a;
+      maxValue = b;
+      break;
+    }
 
-      case PixelFormat_SignedGrayscale16:
-      {
-        int16_t a, b;
-        GetMinMaxValueInternal<int16_t>(a, b, image);
-        minValue = a;
-        maxValue = b;
-        break;
-      }
+    case PixelFormat_SignedGrayscale16:
+    {
+      int16_t a, b;
+      GetMinMaxValueInternal<int16_t>(a, b, image);
+      minValue = a;
+      maxValue = b;
+      break;
+    }
 
-      default:
-        throw OrthancException(ErrorCode_NotImplemented);
+    default:
+      throw OrthancException(ErrorCode_NotImplemented);
     }
   }
 
@@ -946,18 +954,18 @@ namespace Orthanc
   {
     switch (image.GetFormat())
     {
-      case PixelFormat_Float32:
-      {
-        assert(sizeof(float) == 4);
-        float a, b;
-        GetMinMaxValueInternal<float>(a, b, image);
-        minValue = a;
-        maxValue = b;
-        break;
-      }
+    case PixelFormat_Float32:
+    {
+      assert(sizeof(float) == 4);
+      float a, b;
+      GetMinMaxValueInternal<float>(a, b, image);
+      minValue = a;
+      maxValue = b;
+      break;
+    }
 
-      default:
-        throw OrthancException(ErrorCode_NotImplemented);
+    default:
+      throw OrthancException(ErrorCode_NotImplemented);
     }
   }
 
@@ -968,20 +976,20 @@ namespace Orthanc
   {
     switch (image.GetFormat())
     {
-      case PixelFormat_Grayscale8:
-        AddConstantInternal<uint8_t>(image, value);
-        return;
+    case PixelFormat_Grayscale8:
+      AddConstantInternal<uint8_t>(image, value);
+      return;
 
-      case PixelFormat_Grayscale16:
-        AddConstantInternal<uint16_t>(image, value);
-        return;
+    case PixelFormat_Grayscale16:
+      AddConstantInternal<uint16_t>(image, value);
+      return;
 
-      case PixelFormat_SignedGrayscale16:
-        AddConstantInternal<int16_t>(image, value);
-        return;
+    case PixelFormat_SignedGrayscale16:
+      AddConstantInternal<int16_t>(image, value);
+      return;
 
-      default:
-        throw OrthancException(ErrorCode_NotImplemented);
+    default:
+      throw OrthancException(ErrorCode_NotImplemented);
     }
   }
 
@@ -992,41 +1000,41 @@ namespace Orthanc
   {
     switch (image.GetFormat())
     {
-      case PixelFormat_Grayscale8:
-        if (useRound)
-        {
-          MultiplyConstantInternal<uint8_t, true>(image, factor);
-        }
-        else
-        {
-          MultiplyConstantInternal<uint8_t, false>(image, factor);
-        }
-        return;
+    case PixelFormat_Grayscale8:
+      if (useRound)
+      {
+        MultiplyConstantInternal<uint8_t, true>(image, factor);
+      }
+      else
+      {
+        MultiplyConstantInternal<uint8_t, false>(image, factor);
+      }
+      return;
 
-      case PixelFormat_Grayscale16:
-        if (useRound)
-        {
-          MultiplyConstantInternal<uint16_t, true>(image, factor);
-        }
-        else
-        {
-          MultiplyConstantInternal<uint16_t, false>(image, factor);
-        }
-        return;
+    case PixelFormat_Grayscale16:
+      if (useRound)
+      {
+        MultiplyConstantInternal<uint16_t, true>(image, factor);
+      }
+      else
+      {
+        MultiplyConstantInternal<uint16_t, false>(image, factor);
+      }
+      return;
 
-      case PixelFormat_SignedGrayscale16:
-        if (useRound)
-        {
-          MultiplyConstantInternal<int16_t, true>(image, factor);
-        }
-        else
-        {
-          MultiplyConstantInternal<int16_t, false>(image, factor);
-        }
-        return;
+    case PixelFormat_SignedGrayscale16:
+      if (useRound)
+      {
+        MultiplyConstantInternal<int16_t, true>(image, factor);
+      }
+      else
+      {
+        MultiplyConstantInternal<int16_t, false>(image, factor);
+      }
+      return;
 
-      default:
-        throw OrthancException(ErrorCode_NotImplemented);
+    default:
+      throw OrthancException(ErrorCode_NotImplemented);
     }
   }
 
@@ -1038,70 +1046,100 @@ namespace Orthanc
   {
     switch (image.GetFormat())
     {
-      case PixelFormat_Grayscale8:
-        if (useRound)
-        {
-          ShiftScaleInternal<uint8_t, true>(image, offset, scaling);
-        }
-        else
-        {
-          ShiftScaleInternal<uint8_t, false>(image, offset, scaling);
-        }
-        return;
+    case PixelFormat_Grayscale8:
+      if (useRound)
+      {
+        ShiftScaleInternal<uint8_t, true>(image, offset, scaling);
+      }
+      else
+      {
+        ShiftScaleInternal<uint8_t, false>(image, offset, scaling);
+      }
+      return;
 
-      case PixelFormat_Grayscale16:
-        if (useRound)
-        {
-          ShiftScaleInternal<uint16_t, true>(image, offset, scaling);
-        }
-        else
-        {
-          ShiftScaleInternal<uint16_t, false>(image, offset, scaling);
-        }
-        return;
+    case PixelFormat_Grayscale16:
+      if (useRound)
+      {
+        ShiftScaleInternal<uint16_t, true>(image, offset, scaling);
+      }
+      else
+      {
+        ShiftScaleInternal<uint16_t, false>(image, offset, scaling);
+      }
+      return;
 
-      case PixelFormat_SignedGrayscale16:
-        if (useRound)
-        {
-          ShiftScaleInternal<int16_t, true>(image, offset, scaling);
-        }
-        else
-        {
-          ShiftScaleInternal<int16_t, false>(image, offset, scaling);
-        }
-        return;
+    case PixelFormat_SignedGrayscale16:
+      if (useRound)
+      {
+        ShiftScaleInternal<int16_t, true>(image, offset, scaling);
+      }
+      else
+      {
+        ShiftScaleInternal<int16_t, false>(image, offset, scaling);
+      }
+      return;
 
-      default:
-        throw OrthancException(ErrorCode_NotImplemented);
+    default:
+      throw OrthancException(ErrorCode_NotImplemented);
     }
   }
 
 
-  void ImageProcessing::Invert(ImageAccessor& image)
+  void ImageProcessing::Invert(ImageAccessor& image, int64_t maxValue)
   {
     const unsigned int width = image.GetWidth();
     const unsigned int height = image.GetHeight();
-    
+
     switch (image.GetFormat())
     {
-      case PixelFormat_Grayscale8:
-      {
-        for (unsigned int y = 0; y < height; y++)
-        {
-          uint8_t* p = reinterpret_cast<uint8_t*>(image.GetRow(y));
+    case PixelFormat_Grayscale16:
+    {
+      uint16_t maxValueUint16 = (uint16_t)(std::min(maxValue, static_cast<int64_t>(std::numeric_limits<uint16_t>::max())));
 
-          for (unsigned int x = 0; x < width; x++, p++)
-          {
-            *p = 255 - (*p);
-          }
+      for (unsigned int y = 0; y < height; y++)
+      {
+        uint16_t* p = reinterpret_cast<uint16_t*>(image.GetRow(y));
+
+        for (unsigned int x = 0; x < width; x++, p++)
+        {
+          *p = maxValueUint16 - (*p);
         }
-        
-        return;
       }
 
-      default:
-        throw OrthancException(ErrorCode_NotImplemented);
-    }   
+      return;
+    }
+    case PixelFormat_Grayscale8:
+    {
+      uint8_t maxValueUint8 = (uint8_t)(std::min(maxValue, static_cast<int64_t>(std::numeric_limits<uint8_t>::max())));
+
+      for (unsigned int y = 0; y < height; y++)
+      {
+        uint8_t* p = reinterpret_cast<uint8_t*>(image.GetRow(y));
+
+        for (unsigned int x = 0; x < width; x++, p++)
+        {
+          *p = maxValueUint8 - (*p);
+        }
+      }
+
+      return;
+    }
+
+    default:
+      throw OrthancException(ErrorCode_NotImplemented);
+    }
+
+  }
+
+  void ImageProcessing::Invert(ImageAccessor& image)
+  {
+    switch (image.GetFormat())
+    {
+    case PixelFormat_Grayscale8:
+      return Invert(image, 255);
+    default:
+      throw OrthancException(ErrorCode_NotImplemented); // you should use the Invert(image, maxValue) overload
+    }
   }
 
 
@@ -1113,7 +1151,7 @@ namespace Orthanc
     {
     private:
       typedef typename PixelTraits<Format>::PixelType  PixelType;
-    
+
       Orthanc::ImageAccessor&  image_;
       PixelType                value_;
 
@@ -1144,7 +1182,7 @@ namespace Orthanc
             y = y + yi;
             d = d - 2 * dx;
           }
-      
+
           d = d + 2*dy;
         }
       }
@@ -1157,13 +1195,13 @@ namespace Orthanc
         int dx = x1 - x0;
         int dy = y1 - y0;
         int xi = 1;
-    
+
         if (dx < 0)
         {
           xi = -1;
           dx = -dx;
         }
-    
+
         int d = 2 * dx - dy;
         int x = x0;
 
@@ -1176,7 +1214,7 @@ namespace Orthanc
             x = x + xi;
             d = d - 2 * dy;
           }
-      
+
           d = d + 2 * dx;
         }
       }
@@ -1216,7 +1254,7 @@ namespace Orthanc
       {
         // This is an implementation of Bresenham's line algorithm
         // https://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm#All_cases
-    
+
         if (abs(y1 - y0) < abs(x1 - x0))
         {
           if (x0 > x1)
@@ -1252,30 +1290,30 @@ namespace Orthanc
                                         int64_t value)
   {
     switch (image.GetFormat())
-    {       
-      case Orthanc::PixelFormat_Grayscale8:
-      {
-        BresenhamPixelWriter<Orthanc::PixelFormat_Grayscale8> writer(image, value);
-        writer.DrawSegment(x0, y0, x1, y1);
-        break;
-      }
+    {
+    case Orthanc::PixelFormat_Grayscale8:
+    {
+      BresenhamPixelWriter<Orthanc::PixelFormat_Grayscale8> writer(image, value);
+      writer.DrawSegment(x0, y0, x1, y1);
+      break;
+    }
 
-      case Orthanc::PixelFormat_Grayscale16:
-      {
-        BresenhamPixelWriter<Orthanc::PixelFormat_Grayscale16> writer(image, value);
-        writer.DrawSegment(x0, y0, x1, y1);
-        break;
-      }
+    case Orthanc::PixelFormat_Grayscale16:
+    {
+      BresenhamPixelWriter<Orthanc::PixelFormat_Grayscale16> writer(image, value);
+      writer.DrawSegment(x0, y0, x1, y1);
+      break;
+    }
 
-      case Orthanc::PixelFormat_SignedGrayscale16:
-      {
-        BresenhamPixelWriter<Orthanc::PixelFormat_SignedGrayscale16> writer(image, value);
-        writer.DrawSegment(x0, y0, x1, y1);
-        break;
-      }
-        
-      default:
-        throw Orthanc::OrthancException(Orthanc::ErrorCode_NotImplemented);
+    case Orthanc::PixelFormat_SignedGrayscale16:
+    {
+      BresenhamPixelWriter<Orthanc::PixelFormat_SignedGrayscale16> writer(image, value);
+      writer.DrawSegment(x0, y0, x1, y1);
+      break;
+    }
+
+    default:
+      throw Orthanc::OrthancException(Orthanc::ErrorCode_NotImplemented);
     }
   }
 
@@ -1292,46 +1330,186 @@ namespace Orthanc
   {
     switch (image.GetFormat())
     {
-      case Orthanc::PixelFormat_BGRA32:
-      {
-        PixelTraits<Orthanc::PixelFormat_BGRA32>::PixelType pixel;
-        pixel.red_ = red;
-        pixel.green_ = green;
-        pixel.blue_ = blue;
-        pixel.alpha_ = alpha;
+    case Orthanc::PixelFormat_BGRA32:
+    {
+      PixelTraits<Orthanc::PixelFormat_BGRA32>::PixelType pixel;
+      pixel.red_ = red;
+      pixel.green_ = green;
+      pixel.blue_ = blue;
+      pixel.alpha_ = alpha;
 
-        BresenhamPixelWriter<Orthanc::PixelFormat_BGRA32> writer(image, pixel);
-        writer.DrawSegment(x0, y0, x1, y1);
-        break;
-      }
-        
-      case Orthanc::PixelFormat_RGBA32:
-      {
-        PixelTraits<Orthanc::PixelFormat_RGBA32>::PixelType pixel;
-        pixel.red_ = red;
-        pixel.green_ = green;
-        pixel.blue_ = blue;
-        pixel.alpha_ = alpha;
+      BresenhamPixelWriter<Orthanc::PixelFormat_BGRA32> writer(image, pixel);
+      writer.DrawSegment(x0, y0, x1, y1);
+      break;
+    }
 
-        BresenhamPixelWriter<Orthanc::PixelFormat_RGBA32> writer(image, pixel);
-        writer.DrawSegment(x0, y0, x1, y1);
-        break;
-      }
-        
-      case Orthanc::PixelFormat_RGB24:
-      {
-        PixelTraits<Orthanc::PixelFormat_RGB24>::PixelType pixel;
-        pixel.red_ = red;
-        pixel.green_ = green;
-        pixel.blue_ = blue;
+    case Orthanc::PixelFormat_RGBA32:
+    {
+      PixelTraits<Orthanc::PixelFormat_RGBA32>::PixelType pixel;
+      pixel.red_ = red;
+      pixel.green_ = green;
+      pixel.blue_ = blue;
+      pixel.alpha_ = alpha;
 
-        BresenhamPixelWriter<Orthanc::PixelFormat_RGB24> writer(image, pixel);
-        writer.DrawSegment(x0, y0, x1, y1);
-        break;
-      }
-        
-      default:
-        throw Orthanc::OrthancException(Orthanc::ErrorCode_NotImplemented);
+      BresenhamPixelWriter<Orthanc::PixelFormat_RGBA32> writer(image, pixel);
+      writer.DrawSegment(x0, y0, x1, y1);
+      break;
+    }
+
+    case Orthanc::PixelFormat_RGB24:
+    {
+      PixelTraits<Orthanc::PixelFormat_RGB24>::PixelType pixel;
+      pixel.red_ = red;
+      pixel.green_ = green;
+      pixel.blue_ = blue;
+
+      BresenhamPixelWriter<Orthanc::PixelFormat_RGB24> writer(image, pixel);
+      writer.DrawSegment(x0, y0, x1, y1);
+      break;
+    }
+
+    default:
+      throw Orthanc::OrthancException(Orthanc::ErrorCode_NotImplemented);
     }
   }
+
+  void ComputePolygonExtent(int32_t& left, int32_t& right, int32_t& top, int32_t& bottom, const std::vector<ImageProcessing::ImagePoint>& points)
+  {
+    left = std::numeric_limits<int32_t>::max();
+    right = std::numeric_limits<int32_t>::min();
+    top = std::numeric_limits<int32_t>::max();
+    bottom = std::numeric_limits<int32_t>::min();
+
+    for (size_t i = 0; i < points.size(); i++)
+    {
+      const ImageProcessing::ImagePoint& p = points[i];
+      left = std::min(p.GetX(), left);
+      right = std::max(p.GetX(), right);
+      bottom = std::max(p.GetY(), bottom);
+      top = std::min(p.GetY(), top);
+    }
+  }
+
+  template <PixelFormat TargetFormat>
+  void FillPolygon_(ImageAccessor& image,
+              const std::vector<ImageProcessing::ImagePoint>& points,
+              int64_t value_)
+  {
+    typedef typename PixelTraits<TargetFormat>::PixelType  TargetType;
+
+    TargetType value = PixelTraits<TargetFormat>::IntegerToPixel(value_);
+    int32_t left;
+    int32_t right;
+    int32_t top;
+    int32_t bottom;
+
+    ComputePolygonExtent(left, right, top, bottom, points);
+
+    // from http://alienryderflex.com/polygon_fill/
+
+    // convert all "corner"  points to double only once
+    std::vector<double> cpx;
+    std::vector<double> cpy;
+    size_t cpSize = points.size();
+    for (size_t i = 0; i < points.size(); i++)
+    {
+      cpx.push_back((double)points[i].GetX());
+      cpy.push_back((double)points[i].GetY());
+    }
+
+    std::vector<int32_t> nodeX;
+    nodeX.resize(cpSize);
+    int  nodes, pixelX, pixelY, i, j, swap ;
+
+    //  Loop through the rows of the image.
+    for (pixelY = top; pixelY < bottom; pixelY++)
+    {
+      double y = (double)pixelY;
+      //  Build a list of nodes.
+      nodes = 0;
+      j = static_cast<int>(cpSize) - 1;
+
+      for (i = 0; i < static_cast<int>(cpSize); i++)
+      {
+        if ((cpy[i] < y && cpy[j] >=  y) || (cpy[j] < y && cpy[i] >= y))
+        {
+          nodeX[nodes++] = (int32_t)(cpx[i] + (y - cpy[i])/(cpy[j] - cpy[i]) * (cpx[j] - cpx[i]));
+        }
+        j=i;
+      }
+
+      //  Sort the nodes, via a simple “Bubble” sort.
+      i=0;
+      while (i < nodes-1)
+      {
+        if (nodeX[i] > nodeX[i+1])
+        {
+          swap = nodeX[i];
+          nodeX[i] = nodeX[i+1];
+          nodeX[i+1] = swap;
+          if (i > 0)
+          {
+            i--;
+          }
+        }
+        else
+        {
+          i++;
+        }
+      }
+
+      TargetType* row = reinterpret_cast<TargetType*>(image.GetRow(pixelY));
+      //  Fill the pixels between node pairs.
+      for (i=0; i<nodes; i+=2)
+      {
+        if (nodeX[i] >= right)
+          break;
+
+        if (nodeX[i+1] >= left)
+        {
+          if (nodeX[i]< left)
+          {
+            nodeX[i] = left;
+          }
+
+          if (nodeX[i+1] > right)
+          {
+            nodeX[i+1] = right;
+          }
+
+          for (pixelX = nodeX[i]; pixelX <= nodeX[i+1]; pixelX++)
+          {
+            *(row + pixelX) = value;
+          }
+        }
+      }
+    }
+  }
+
+  void ImageProcessing::FillPolygon(ImageAccessor& image,
+                                    const std::vector<ImagePoint>& points,
+                                    int64_t value)
+  {
+    switch (image.GetFormat())
+    {
+    case Orthanc::PixelFormat_Grayscale8:
+    {
+      FillPolygon_<Orthanc::PixelFormat_Grayscale8>(image, points, value);
+      break;
+    }
+    case Orthanc::PixelFormat_Grayscale16:
+    {
+      FillPolygon_<Orthanc::PixelFormat_Grayscale16>(image, points, value);
+      break;
+    }
+    case Orthanc::PixelFormat_SignedGrayscale16:
+    {
+      FillPolygon_<Orthanc::PixelFormat_SignedGrayscale16>(image, points, value);
+      break;
+    }
+    default:
+      throw Orthanc::OrthancException(Orthanc::ErrorCode_NotImplemented);
+    }
+  }
+
 }
