@@ -2,7 +2,7 @@
  * Orthanc - A Lightweight, RESTful DICOM Store
  * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
  * Department, University Hospital of Liege, Belgium
- * Copyright (C) 2017 Osimis, Belgium
+ * Copyright (C) 2017-2019 Osimis S.A., Belgium
  *
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -197,12 +197,12 @@ namespace Orthanc
         Toolbox::TokenizeString(accepted, it->second, ';');
         for (size_t i = 0; i < accepted.size(); i++)
         {
-          if (accepted[i] == "application/xml")
+          if (accepted[i] == MIME_XML)
           {
             wrappedOutput.SetConvertJsonToXml(true);
           }
 
-          if (accepted[i] == "application/json")
+          if (accepted[i] == MIME_JSON)
           {
             wrappedOutput.SetConvertJsonToXml(false);
           }
