@@ -13,6 +13,8 @@ if ($.browser.msie)
 //$.mobile.defaultPageTransition = 'slide';
 
 
+var LIMIT_RESOURCES = 100;
+
 var currentPage = '';
 var currentUuid = '';
 
@@ -26,9 +28,11 @@ function DeepCopy(obj)
 function ChangePage(page, options)
 {
   var first = true;
+  var value;
+
   if (options) {
     for (var key in options) {
-      var value = options[key];
+      value = options[key];
       if (first) {
         page += '?';
         first = false;
@@ -224,11 +228,13 @@ function CompleteFormatting(node, link, isReverse, count)
 
 function FormatMainDicomTags(target, tags, tagsToIgnore)
 {
+  var v;
+  
   for (var i in tags)
   {
     if (tagsToIgnore.indexOf(i) == -1)
     {
-      var v = tags[i];
+      v = tags[i];
 
       if (i == "PatientBirthDate" ||
           i == "StudyDate" ||
@@ -264,15 +270,30 @@ function FormatPatient(patient, link, isReverse)
 
 
 
-function FormatStudy(study, link, isReverse)
+function FormatStudy(study, link, isReverse, includePatient)
 {
-  var node = $('<div>').append($('<h3>').text(study.MainDicomTags.StudyDescription));
+  var label;
+  var node;
 
-  FormatMainDicomTags(node, study.MainDicomTags, [ 
-     "StudyDescription", 
-     "StudyTime" 
-  ]);
+  if (includePatient) {
+    label = study.Label;
+  } else {
+    label = study.MainDicomTags.StudyDescription;
+  }
+
+  node = $('<div>').append($('<h3>').text(label));
+
+  if (includePatient) {
+    FormatMainDicomTags(node, study.PatientMainDicomTags, [ 
+      'PatientName'
+    ]);
+  }
     
+  FormatMainDicomTags(node, study.MainDicomTags, [ 
+     'StudyDescription', 
+     'StudyTime' 
+  ]);
+
   return CompleteFormatting(node, link, isReverse, study.Series.length);
 }
 
@@ -281,6 +302,8 @@ function FormatStudy(study, link, isReverse)
 function FormatSeries(series, link, isReverse)
 {
   var c;
+  var node;
+
   if (series.ExpectedNumberOfInstances == null ||
       series.Instances.length == series.ExpectedNumberOfInstances)
   {
@@ -291,7 +314,7 @@ function FormatSeries(series, link, isReverse)
     c = series.Instances.length + '/' + series.ExpectedNumberOfInstances;
   }
   
-  var node = $('<div>')
+  node = $('<div>')
       .append($('<h3>').text(series.MainDicomTags.SeriesDescription))
       .append($('<p>').append($('<em>')
                            .text('Status: ')
@@ -346,19 +369,171 @@ $('[data-role="page"]').live('pagebeforeshow', function() {
 
 
 
-$('#find-patients').live('pagebeforeshow', function() {
-  GetResource('/patients?expand', function(patients) {
-      var target = $('#all-patients');
-      $('li', target).remove();
-    
-      SortOnDicomTag(patients, 'PatientName', false, false);
+$('#lookup').live('pagebeforeshow', function() {
+  // NB: "GenerateDicomDate()" is defined in "query-retrieve.js"
+  var target = $('#lookup-study-date');
+  $('option', target).remove();
+  target.append($('<option>').attr('value', '*').text('Any date'));
+  target.append($('<option>').attr('value', GenerateDicomDate(0)).text('Today'));
+  target.append($('<option>').attr('value', GenerateDicomDate(-1)).text('Yesterday'));
+  target.append($('<option>').attr('value', GenerateDicomDate(-7) + '-').text('Last 7 days'));
+  target.append($('<option>').attr('value', GenerateDicomDate(-31) + '-').text('Last 31 days'));
+  target.append($('<option>').attr('value', GenerateDicomDate(-31 * 3) + '-').text('Last 3 months'));
+  target.append($('<option>').attr('value', GenerateDicomDate(-365) + '-').text('Last year'));
+  target.selectmenu('refresh');
 
-      for (var i = 0; i < patients.length; i++) {
-        var p = FormatPatient(patients[i], '#patient?uuid=' + patients[i].ID);
-        target.append(p);
+  $('#lookup-result').hide();
+});
+
+
+$('#lookup-submit').live('click', function() {
+  var lookup;
+
+  $('#lookup-result').hide();
+
+  lookup = {
+    'Level' : 'Study',
+    'Expand' : true,
+    'Limit' : LIMIT_RESOURCES + 1,
+    'Query' : {
+      'StudyDate' : $('#lookup-study-date').val()
+    }
+  };
+
+  $('#lookup-form input').each(function(index, input) {
+    if (input.value.length != 0) {
+      if (input.id == 'lookup-patient-id') {
+        lookup['Query']['PatientID'] = input.value;
+      } 
+      else if (input.id == 'lookup-patient-name') {
+        lookup['Query']['PatientName'] = input.value;
+      } 
+      else if (input.id == 'lookup-accession-number') {
+        lookup['Query']['AccessionNumber'] = input.value;
+      } 
+      else if (input.id == 'lookup-study-description') {
+        lookup['Query']['StudyDescription'] = input.value;
+      }
+      else {
+        console.error('Unknown lookup field: ' + input.id);
+      }
+    } 
+  });
+
+  $.ajax({
+    url: '../tools/find',
+    type: 'POST', 
+    data: JSON.stringify(lookup),
+    dataType: 'json',
+    async: false,
+    error: function() {
+      alert('Error during lookup');
+    },
+    success: function(studies) {
+      FormatListOfStudies('#lookup-result ul', '#lookup-alert', '#lookup-count', studies);
+      $('#lookup-result').show();
+    }
+  });
+
+  return false;
+});
+
+
+$('#find-patients').live('pagebeforeshow', function() {
+  GetResource('/patients?expand&since=0&limit=' + (LIMIT_RESOURCES + 1), function(patients) {
+    var target = $('#all-patients');
+    var count, showAlert, p;
+
+
+    $('li', target).remove();
+    
+    SortOnDicomTag(patients, 'PatientName', false, false);
+
+    if (patients.length <= LIMIT_RESOURCES) {
+      count = patients.length;
+      showAlert = false;
+    }
+    else {
+      count = LIMIT_RESOURCES;
+      showAlert = true;
+    }
+
+    for (var i = 0; i < count; i++) {
+      p = FormatPatient(patients[i], '#patient?uuid=' + patients[i].ID);
+      target.append(p);
+    }
+
+    target.listview('refresh'); 
+
+    if (showAlert) {
+      $('#count-patients').text(LIMIT_RESOURCES);
+      $('#alert-patients').show();
+    } else {
+      $('#alert-patients').hide();
+    }
+  });
+});
+
+
+
+function FormatListOfStudies(targetId, alertId, countId, studies)
+{
+  var target = $(targetId);
+  var patient, study, s;
+  var count, showAlert;
+
+  $('li', target).remove();
+
+  for (var i = 0; i < studies.length; i++) {
+    patient = studies[i].PatientMainDicomTags.PatientName;
+    study = studies[i].MainDicomTags.StudyDescription;
+
+    s = "";
+    if (typeof patient === 'string') {
+      s = patient;
+    }
+
+    if (typeof study === 'string') {
+      if (s.length > 0) {
+        s += ' - ';
       }
 
-      target.listview('refresh');
+      s += study;
+    }
+
+    studies[i]['Label'] = s;
+  }
+
+  Sort(studies, function(a) { return a.Label }, false, false);
+
+  if (studies.length <= LIMIT_RESOURCES) {
+    count = studies.length;
+    showAlert = false;
+  }
+  else {
+    count = LIMIT_RESOURCES;
+    showAlert = true;
+  }
+
+  for (var i = 0; i < count; i++) {
+    s = FormatStudy(studies[i], '#study?uuid=' + studies[i].ID, false, true);
+    target.append(s);
+  }
+
+  target.listview('refresh');
+
+  if (showAlert) {
+    $(countId).text(LIMIT_RESOURCES);
+    $(alertId).show();
+  } else {
+    $(alertId).hide();
+  }
+}
+
+
+$('#find-studies').live('pagebeforeshow', function() {
+  GetResource('/studies?expand&since=0&limit=' + (LIMIT_RESOURCES + 1), function(studies) {
+    FormatListOfStudies('#all-studies', '#alert-studies', '#count-studies', studies);
   });
 });
 
@@ -383,8 +558,10 @@ function SetupAnonymizedOrModifiedFrom(buttonSelector, resource, resourceType, f
 
 function RefreshPatient()
 {
+  var pageData, target, v;
+
   if ($.mobile.pageData) {
-    var pageData = DeepCopy($.mobile.pageData);
+    pageData = DeepCopy($.mobile.pageData);
 
     GetResource('/patients/' + pageData.uuid, function(patient) {
       GetResource('/patients/' + pageData.uuid + '/studies', function(studies) {
@@ -396,7 +573,7 @@ function RefreshPatient()
           .append(FormatPatient(patient))
           .listview('refresh');
 
-        var target = $('#list-studies');
+        target = $('#list-studies');
         $('li', target).remove();
         
         for (var i = 0; i < studies.length; i++) {
@@ -423,7 +600,7 @@ function RefreshPatient()
           async: false,
           cache: false,
           success: function (s) {
-            var v = (s == '1') ? 'on' : 'off';
+            v = (s == '1') ? 'on' : 'off';
             $('#protection').val(v).slider('refresh');
           }
         });
@@ -438,8 +615,10 @@ function RefreshPatient()
 
 function RefreshStudy()
 {
+  var pageData, target;
+
   if ($.mobile.pageData) {
-    var pageData = DeepCopy($.mobile.pageData);
+    pageData = DeepCopy($.mobile.pageData);
 
     GetResource('/studies/' + pageData.uuid, function(study) {
       GetResource('/patients/' + study.ParentPatient, function(patient) {
@@ -458,7 +637,7 @@ function RefreshStudy()
           SetupAnonymizedOrModifiedFrom('#study-anonymized-from', study, 'study', 'AnonymizedFrom');
           SetupAnonymizedOrModifiedFrom('#study-modified-from', study, 'study', 'ModifiedFrom');
 
-          var target = $('#list-series');
+          target = $('#list-series');
           $('li', target).remove();
           for (var i = 0; i < series.length; i++) {
             if (i == 0 || series[i].MainDicomTags.SeriesDate != series[i - 1].MainDicomTags.SeriesDate)
@@ -483,8 +662,10 @@ function RefreshStudy()
 
 function RefreshSeries() 
 {
+  var pageData, target;
+
   if ($.mobile.pageData) {
-    var pageData = DeepCopy($.mobile.pageData);
+    pageData = DeepCopy($.mobile.pageData);
 
     GetResource('/series/' + pageData.uuid, function(series) {
       GetResource('/studies/' + series.ParentStudy, function(study) {
@@ -508,7 +689,7 @@ function RefreshSeries()
             SetupAnonymizedOrModifiedFrom('#series-anonymized-from', series, 'series', 'AnonymizedFrom');
             SetupAnonymizedOrModifiedFrom('#series-modified-from', series, 'series', 'ModifiedFrom');
 
-            var target = $('#list-instances');
+            target = $('#list-instances');
             $('li', target).remove();
             for (var i = 0; i < instances.length; i++) {
               target.append(FormatInstance(instances[i], '#instance?uuid=' + instances[i].ID));
@@ -547,10 +728,11 @@ function EscapeHtml(value)
 function ConvertForTree(dicom)
 {
   var result = [];
+  var label, c;
 
   for (var i in dicom) {
     if (dicom[i] != null) {
-      var label = (i + '<span class="tag-name"> (<i>' +
+      label = (i + '<span class="tag-name"> (<i>' +
                    EscapeHtml(dicom[i]["Name"]) +
                    '</i>)</span>: ');
 
@@ -577,7 +759,7 @@ function ConvertForTree(dicom)
       }
       else if (dicom[i]["Type"] == 'Sequence')
       {
-        var c = [];
+        c = [];
         for (var j = 0; j < dicom[i]["Value"].length; j++) {
           c.push({
             label: 'Item ' + j,
@@ -599,8 +781,10 @@ function ConvertForTree(dicom)
 
 function RefreshInstance()
 {
+  var pageData;
+
   if ($.mobile.pageData) {
-    var pageData = DeepCopy($.mobile.pageData);
+    pageData = DeepCopy($.mobile.pageData);
 
     GetResource('/instances/' + pageData.uuid, function(instance) {
       GetResource('/series/' + instance.ParentSeries, function(series) {
@@ -676,7 +860,7 @@ function DeleteResource(path)
     success: function(s) {
       var ancestor = s.RemainingAncestor;
       if (ancestor == null)
-        $.mobile.changePage('#find-patients');
+        $.mobile.changePage('#lookup');
       else
         $.mobile.changePage('#' + ancestor.Type.toLowerCase() + '?uuid=' + ancestor.ID);
     }
@@ -749,10 +933,12 @@ $('#instance-download-json').live('click', function(e) {
 
 
 $('#instance-preview').live('click', function(e) {
-  if ($.mobile.pageData) {
-    var pageData = DeepCopy($.mobile.pageData);
+  var pageData, pdf, images;
 
-    var pdf = '../instances/' + pageData.uuid + '/pdf';
+  if ($.mobile.pageData) {
+    pageData = DeepCopy($.mobile.pageData);
+
+    pdf = '../instances/' + pageData.uuid + '/pdf';
     $.ajax({
       url: pdf,
       cache: false,
@@ -774,7 +960,7 @@ $('#instance-preview').live('click', function(e) {
           {
             // Viewing a multi-frame image
 
-            var images = [];
+            images = [];
             for (var i = 0; i < frames.length; i++) {
               images.push([ '../instances/' + pageData.uuid + '/frames/' + i + '/preview' ]);
             }
@@ -795,14 +981,16 @@ $('#instance-preview').live('click', function(e) {
 
 
 $('#series-preview').live('click', function(e) {
+  var pageData, images;
+
   if ($.mobile.pageData) {
-    var pageData = DeepCopy($.mobile.pageData);
+    pageData = DeepCopy($.mobile.pageData);
 
     GetResource('/series/' + pageData.uuid, function(series) {
       GetResource('/series/' + pageData.uuid + '/instances', function(instances) {
         Sort(instances, function(x) { return x.IndexInSeries; }, true, false);
 
-        var images = [];
+        images = [];
         for (var i = 0; i < instances.length; i++) {
           images.push([ '../instances/' + instances[i].ID + '/preview',
                         (i + 1).toString() + '/' + instances.length.toString() ])
@@ -839,13 +1027,15 @@ function ChooseDicomModality(callback)
     async: false,
     cache: false,
     success: function(modalities) {
+      var name, item;
+      
       if (modalities.length > 0)
       {
         items.append('<li data-role="list-divider">DICOM modalities</li>');
 
         for (var i = 0; i < modalities.length; i++) {
-          var name = modalities[i];
-          var item = $('<li>')
+          name = modalities[i];
+          item = $('<li>')
             .html('<a href="#" rel="close">' + name + '</a>')
             .attr('name', name)
             .click(function() { 
@@ -863,13 +1053,15 @@ function ChooseDicomModality(callback)
         async: false,
         cache: false,
         success: function(peers) {
+          var name, item;
+
           if (peers.length > 0)
           {
             items.append('<li data-role="list-divider">Orthanc peers</li>');
 
             for (var i = 0; i < peers.length; i++) {
-              var name = peers[i];
-              var item = $('<li>')
+              name = peers[i];
+              item = $('<li>')
                 .html('<a href="#" rel="close">' + name + '</a>')
                 .attr('name', name)
                 .click(function() { 
@@ -909,9 +1101,7 @@ function ChooseDicomModality(callback)
 $('#instance-store,#series-store,#study-store,#patient-store').live('click', function(e) {
   ChooseDicomModality(function(modality, peer) {
     var pageData = DeepCopy($.mobile.pageData);
-
-    var url;
-    var loading;
+    var url, loading;
 
     if (modality != '')
     {
@@ -1104,3 +1294,220 @@ $('#plugins').live('pagebeforeshow', function() {
     }
   });
 });
+
+
+
+function ParseJobTime(s)
+{
+  var t = (s.substr(0, 4) + '-' +
+           s.substr(4, 2) + '-' +
+           s.substr(6, 5) + ':' +
+           s.substr(11, 2) + ':' +
+           s.substr(13));
+  var utc = new Date(t);
+
+  // Convert from UTC to local time
+  return new Date(utc.getTime() - utc.getTimezoneOffset() * 60000);
+}
+
+
+function AddJobField(target, description, field)
+{
+  if (!(typeof field === 'undefined')) {
+    target.append($('<p>')
+                  .text(description)
+                  .append($('<strong>').text(field)));
+  }
+}
+
+
+function AddJobDateField(target, description, field)
+{
+  if (!(typeof field === 'undefined')) {
+    target.append($('<p>')
+                  .text(description)
+                  .append($('<strong>').text(ParseJobTime(field))));
+  }
+}
+
+
+$('#jobs').live('pagebeforeshow', function() {
+  $.ajax({
+    url: '../jobs?expand',
+    dataType: 'json',
+    async: false,
+    cache: false,
+    success: function(jobs) {
+      var target = $('#all-jobs');
+      var running, pending, inactive;
+
+      $('li', target).remove();
+
+      running = $('<li>')
+          .attr('data-role', 'list-divider')
+          .text('Currently running');
+
+      pending = $('<li>')
+          .attr('data-role', 'list-divider')
+          .text('Pending jobs');
+
+      inactive = $('<li>')
+          .attr('data-role', 'list-divider')
+          .text('Inactive jobs');
+
+      target.append(running);
+      target.append(pending);
+      target.append(inactive);
+
+      jobs.map(function(job) {
+        var li = $('<li>');
+        var item = $('<a>');
+        
+        li.append(item);
+        item.attr('href', '#job?uuid=' + job.ID);
+        item.append($('<h1>').text(job.Type));
+        item.append($('<span>').addClass('ui-li-count').text(job.State));
+        AddJobField(item, 'ID: ', job.ID);
+        AddJobField(item, 'Local AET: ', job.Content.LocalAet);
+        AddJobField(item, 'Remote AET: ', job.Content.RemoteAet);
+        AddJobDateField(item, 'Creation time: ', job.CreationTime);
+        AddJobDateField(item, 'Completion time: ', job.CompletionTime);
+        AddJobDateField(item, 'ETA: ', job.EstimatedTimeOfArrival);
+
+        if (job.State == 'Running' ||
+            job.State == 'Pending' ||
+            job.State == 'Paused') {
+          AddJobField(item, 'Priority: ', job.Priority);
+          AddJobField(item, 'Progress: ', job.Progress);
+        }
+        
+        if (job.State == 'Running') {
+          li.insertAfter(running);
+        } else if (job.State == 'Pending' ||
+                   job.State == 'Paused') {
+          li.insertAfter(pending);
+        } else {
+          li.insertAfter(inactive);
+        }
+      });
+
+      target.listview('refresh');
+    }
+  });
+});
+
+
+$('#job').live('pagebeforeshow', function() {
+  var pageData, target;
+
+  if ($.mobile.pageData) {
+    pageData = DeepCopy($.mobile.pageData);
+
+    $.ajax({
+      url: '../jobs/' + pageData.uuid,
+      dataType: 'json',
+      async: false,
+      cache: false,
+      success: function(job) {
+        var block, value;
+        
+        target = $('#job-info');
+        $('li', target).remove();
+
+        target.append($('<li>')
+                      .attr('data-role', 'list-divider')
+                      .text('General information about the job'));
+
+        {                       
+          block = $('<li>');
+          for (var i in job) {
+            if (i == 'CreationTime' ||
+                i == 'CompletionTime' ||
+                i == 'EstimatedTimeOfArrival') {
+              AddJobDateField(block, i + ': ', job[i]);
+            } else if (i != 'InternalContent' &&
+                      i != 'Content' &&
+                      i != 'Timestamp') {
+              AddJobField(block, i + ': ', job[i]);
+            }
+          }
+        }
+
+        target.append(block);
+        
+        target.append($('<li>')
+                      .attr('data-role', 'list-divider')
+                      .text('Detailed information'));
+
+        {
+          block = $('<li>');
+
+          for (var item in job.Content) {
+            var value = job.Content[item];
+            if (typeof value !== 'string') {
+              value = JSON.stringify(value);
+            }
+            
+            AddJobField(block, item + ': ', value);
+          }
+        }
+        
+        target.append(block);
+        
+        target.listview('refresh');
+
+        $('#job-cancel').closest('.ui-btn').hide();
+        $('#job-retry').closest('.ui-btn').hide();
+        $('#job-resubmit').closest('.ui-btn').hide();
+        $('#job-pause').closest('.ui-btn').hide();
+        $('#job-resume').closest('.ui-btn').hide();
+
+        if (job.State == 'Running' ||
+            job.State == 'Pending' ||
+            job.State == 'Retry') {
+          $('#job-cancel').closest('.ui-btn').show();
+          $('#job-pause').closest('.ui-btn').show();
+        }
+        else if (job.State == 'Success') {
+        }
+        else if (job.State == 'Failure') {
+          $('#job-resubmit').closest('.ui-btn').show();
+        }
+        else if (job.State == 'Paused') {
+          $('#job-resume').closest('.ui-btn').show();
+        }
+      }
+    });
+  }
+});
+
+
+
+function TriggerJobAction(action)
+{
+  $.ajax({
+    url: '../jobs/' + $.mobile.pageData.uuid + '/' + action,
+    type: 'POST',
+    async: false,
+    cache: false,
+    complete: function(s) {
+      window.location.reload();
+    }
+  });
+}
+
+$('#job-cancel').live('click', function() {
+  TriggerJobAction('cancel');
+});
+
+$('#job-resubmit').live('click', function() {
+  TriggerJobAction('resubmit');
+});
+
+$('#job-pause').live('click', function() {
+  TriggerJobAction('pause');
+});
+
+$('#job-resume').live('click', function() {
+  TriggerJobAction('resume');
+});
