@@ -2,7 +2,7 @@
  * Orthanc - A Lightweight, RESTful DICOM Store
  * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
  * Department, University Hospital of Liege, Belgium
- * Copyright (C) 2017 Osimis, Belgium
+ * Copyright (C) 2017-2019 Osimis S.A., Belgium
  *
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -34,10 +34,13 @@
 #include "../PrecompiledHeadersServer.h"
 #include "OrthancRestApi.h"
 
-#include "../../Core/Logging.h"
 #include "../../Core/DicomParsing/FromDcmtkBridge.h"
+#include "../../Core/Logging.h"
+#include "../../Core/SerializationToolbox.h"
 #include "../ServerContext.h"
-#include "../OrthancInitialization.h"
+#include "../ServerJobs/MergeStudyJob.h"
+#include "../ServerJobs/ResourceModificationJob.h"
+#include "../ServerJobs/SplitStudyJob.h"
 
 #include <boost/lexical_cast.hpp>
 #include <boost/algorithm/string/predicate.hpp>
@@ -46,95 +49,7 @@ namespace Orthanc
 {
   // Modification of DICOM instances ------------------------------------------
 
-  enum TagOperation
-  {
-    TagOperation_Keep,
-    TagOperation_Remove
-  };
-
-  static bool IsDatabaseKey(const DicomTag& tag)
-  {
-    return (tag == DICOM_TAG_PATIENT_ID ||
-            tag == DICOM_TAG_STUDY_INSTANCE_UID ||
-            tag == DICOM_TAG_SERIES_INSTANCE_UID ||
-            tag == DICOM_TAG_SOP_INSTANCE_UID);
-  }
-
-  static void ParseListOfTags(DicomModification& target,
-                              const Json::Value& query,
-                              TagOperation operation,
-                              bool force)
-  {
-    if (!query.isArray())
-    {
-      throw OrthancException(ErrorCode_BadRequest);
-    }
-
-    for (Json::Value::ArrayIndex i = 0; i < query.size(); i++)
-    {
-      std::string name = query[i].asString();
-
-      DicomTag tag = FromDcmtkBridge::ParseTag(name);
-
-      if (!force && IsDatabaseKey(tag))
-      {
-        LOG(ERROR) << "Marking tag \"" << name << "\" as to be "
-                   << (operation == TagOperation_Keep ? "kept" : "removed")
-                   << " requires the \"Force\" option to be set to true";
-        throw OrthancException(ErrorCode_BadRequest);
-      }
-
-      switch (operation)
-      {
-        case TagOperation_Keep:
-          target.Keep(tag);
-          VLOG(1) << "Keep: " << name << " " << tag << std::endl;
-          break;
-
-        case TagOperation_Remove:
-          target.Remove(tag);
-          VLOG(1) << "Remove: " << name << " " << tag << std::endl;
-          break;
-
-        default:
-          throw OrthancException(ErrorCode_InternalError);
-      }
-    }
-  }
-
-
-  static void ParseReplacements(DicomModification& target,
-                                const Json::Value& replacements,
-                                bool force)
-  {
-    if (!replacements.isObject())
-    {
-      throw OrthancException(ErrorCode_BadRequest);
-    }
-
-    Json::Value::Members members = replacements.getMemberNames();
-    for (size_t i = 0; i < members.size(); i++)
-    {
-      const std::string& name = members[i];
-      const Json::Value& value = replacements[name];
-
-      DicomTag tag = FromDcmtkBridge::ParseTag(name);
-
-      if (!force && IsDatabaseKey(tag))
-      {
-        LOG(ERROR) << "Marking tag \"" << name << "\" as to be replaced "
-                   << "requires the \"Force\" option to be set to true";
-        throw OrthancException(ErrorCode_BadRequest);
-      }
-
-      target.Replace(tag, value, false);
-
-      VLOG(1) << "Replace: " << name << " " << tag 
-              << " == " << value.toStyledString() << std::endl;
-    }
-  }
-
-
+  
   static std::string GeneratePatientName(ServerContext& context)
   {
     uint64_t seq = context.GetIndex().IncrementGlobalSequence(GlobalProperty_AnonymizationSequence);
@@ -142,145 +57,46 @@ namespace Orthanc
   }
 
 
-  static bool GetBooleanValue(const std::string& member,
-                              const Json::Value& json,
-                              bool defaultValue)
+  static void ParseModifyRequest(Json::Value& request,
+                                 DicomModification& target,
+                                 const RestApiPostCall& call)
   {
-    if (!json.isMember(member))
+    // curl http://localhost:8042/series/95a6e2bf-9296e2cc-bf614e2f-22b391ee-16e010e0/modify -X POST -d '{"Replace":{"InstitutionName":"My own clinic"},"Priority":9}'
+
+    if (call.ParseJsonRequest(request))
     {
-      return defaultValue;
-    }
-    else if (json[member].type() == Json::booleanValue)
-    {
-      return json[member].asBool();
+      target.ParseModifyRequest(request);
     }
     else
     {
-      LOG(ERROR) << "Member \"" << member << "\" should be a Boolean value";
       throw OrthancException(ErrorCode_BadFileFormat);
     }
   }
 
 
-  bool OrthancRestApi::ParseModifyRequest(DicomModification& target,
-                                          const Json::Value& request)
-  {
-    if (request.isObject())
-    {
-      bool force = GetBooleanValue("Force", request, false);
-      
-      if (GetBooleanValue("RemovePrivateTags", request, false))
-      {
-        target.SetRemovePrivateTags(true);
-      }
-
-      if (request.isMember("Remove"))
-      {
-        ParseListOfTags(target, request["Remove"], TagOperation_Remove, force);
-      }
-
-      if (request.isMember("Replace"))
-      {
-        ParseReplacements(target, request["Replace"], force);
-      }
-
-      // The "Keep" operation only makes sense for the tags
-      // StudyInstanceUID, SeriesInstanceUID and SOPInstanceUID. Avoid
-      // this feature as much as possible, as this breaks the DICOM
-      // model of the real world, except if you know exactly what
-      // you're doing!
-      if (request.isMember("Keep"))
-      {
-        ParseListOfTags(target, request["Keep"], TagOperation_Keep, force);
-      }
-
-      return true;
-    }
-    else
-    {
-      return false;
-    }
-  }
-
-
-  static bool ParseModifyRequest(DicomModification& target,
-                                 const RestApiPostCall& call)
-  {
-    // curl http://localhost:8042/series/95a6e2bf-9296e2cc-bf614e2f-22b391ee-16e010e0/modify -X POST -d '{"Replace":{"InstitutionName":"My own clinic"}}'
-
-    Json::Value request;
-    if (call.ParseJsonRequest(request))
-    {
-      return OrthancRestApi::ParseModifyRequest(target, request);
-    }
-    else
-    {
-      return false;
-    }
-  }
-
-
-  static bool ParseAnonymizationRequest(DicomModification& target,
+  static void ParseAnonymizationRequest(Json::Value& request,
+                                        DicomModification& target,
                                         RestApiPostCall& call)
   {
     // curl http://localhost:8042/instances/6e67da51-d119d6ae-c5667437-87b9a8a5-0f07c49f/anonymize -X POST -d '{"Replace":{"PatientName":"hello","0010-0020":"world"},"Keep":["StudyDescription", "SeriesDescription"],"KeepPrivateTags": true,"Remove":["Modality"]}' > Anonymized.dcm
 
-    Json::Value request;
-    if (!call.ParseJsonRequest(request) ||
-        !request.isObject())
+    if (call.ParseJsonRequest(request) &&
+        request.isObject())
     {
-      return false;
-    }
+      bool patientNameReplaced;
+      target.ParseAnonymizationRequest(patientNameReplaced, request);
 
-    bool force = GetBooleanValue("Force", request, false);
-      
-    // As of Orthanc 1.3.0, the default anonymization is done
-    // according to PS 3.15-2017c Table E.1-1 (basic profile)
-    DicomVersion version = DicomVersion_2017c;
-    if (request.isMember("DicomVersion"))
-    {
-      if (request["DicomVersion"].type() != Json::stringValue)
+      if (patientNameReplaced)
       {
-        throw OrthancException(ErrorCode_BadFileFormat);
-      }
-      else
-      {
-        version = StringToDicomVersion(request["DicomVersion"].asString());
+        // Overwrite the random Patient's Name by one that is more
+        // user-friendly (provided none was specified by the user)
+        target.Replace(DICOM_TAG_PATIENT_NAME, GeneratePatientName(OrthancRestApi::GetContext(call)), true);
       }
     }
-        
-    target.SetupAnonymization(version);
-    std::string patientName = target.GetReplacementAsString(DICOM_TAG_PATIENT_NAME);
-
-    if (GetBooleanValue("KeepPrivateTags", request, false))
+    else
     {
-      target.SetRemovePrivateTags(false);
+      throw OrthancException(ErrorCode_BadFileFormat);
     }
-
-    if (request.isMember("Remove"))
-    {
-      ParseListOfTags(target, request["Remove"], TagOperation_Remove, force);
-    }
-
-    if (request.isMember("Replace"))
-    {
-      ParseReplacements(target, request["Replace"], force);
-    }
-
-    if (request.isMember("Keep"))
-    {
-      ParseListOfTags(target, request["Keep"], TagOperation_Keep, force);
-    }
-
-    if (target.IsReplaced(DICOM_TAG_PATIENT_NAME) &&
-        target.GetReplacement(DICOM_TAG_PATIENT_NAME) == patientName)
-    {
-      // Overwrite the random Patient's Name by one that is more
-      // user-friendly (provided none was specified by the user)
-      target.Replace(DICOM_TAG_PATIENT_NAME, GeneratePatientName(OrthancRestApi::GetContext(call)), true);
-    }
-
-    return true;
   }
 
 
@@ -289,180 +105,44 @@ namespace Orthanc
   {
     std::string id = call.GetUriComponent("id", "");
 
-    ServerContext::DicomCacheLocker locker(OrthancRestApi::GetContext(call), id);
+    std::auto_ptr<ParsedDicomFile> modified;
 
-    std::auto_ptr<ParsedDicomFile> modified(locker.GetDicom().Clone());
+    {
+      ServerContext::DicomCacheLocker locker(OrthancRestApi::GetContext(call), id);
+      modified.reset(locker.GetDicom().Clone(true));
+    }
+    
     modification.Apply(*modified);
     modified->Answer(call.GetOutput());
   }
 
 
-  static void AnonymizeOrModifyResource(DicomModification& modification,
-                                        MetadataType metadataType,
-                                        ChangeType changeType,
-                                        ResourceType resourceType,
-                                        RestApiPostCall& call)
-  {
-    bool isFirst = true;
-    Json::Value result(Json::objectValue);
-
-    ServerContext& context = OrthancRestApi::GetContext(call);
-
-    typedef std::list<std::string> Instances;
-    Instances instances;
-    std::string id = call.GetUriComponent("id", "");
-    context.GetIndex().GetChildInstances(instances, id);
-
-    if (instances.empty())
-    {
-      return;
-    }
-
-
-    /**
-     * Loop over all the instances of the resource.
-     **/
-
-    for (Instances::const_iterator it = instances.begin(); 
-         it != instances.end(); ++it)
-    {
-      LOG(INFO) << "Modifying instance " << *it;
-
-      std::auto_ptr<ServerContext::DicomCacheLocker> locker;
-
-      try
-      {
-        locker.reset(new ServerContext::DicomCacheLocker(OrthancRestApi::GetContext(call), *it));
-      }
-      catch (OrthancException&)
-      {
-        // This child instance has been removed in between
-        continue;
-      }
-
-
-      ParsedDicomFile& original = locker->GetDicom();
-      DicomInstanceHasher originalHasher = original.GetHasher();
-
-
-      /**
-       * Compute the resulting DICOM instance.
-       **/
-
-      std::auto_ptr<ParsedDicomFile> modified(original.Clone());
-      modification.Apply(*modified);
-
-      DicomInstanceToStore toStore;
-      toStore.SetRestOrigin(call);
-      toStore.SetParsedDicomFile(*modified);
-
-
-      /**
-       * Prepare the metadata information to associate with the
-       * resulting DICOM instance (AnonymizedFrom/ModifiedFrom).
-       **/
-
-      DicomInstanceHasher modifiedHasher = modified->GetHasher();
-
-      if (originalHasher.HashSeries() != modifiedHasher.HashSeries())
-      {
-        toStore.AddMetadata(ResourceType_Series, metadataType, originalHasher.HashSeries());
-      }
-
-      if (originalHasher.HashStudy() != modifiedHasher.HashStudy())
-      {
-        toStore.AddMetadata(ResourceType_Study, metadataType, originalHasher.HashStudy());
-      }
-
-      if (originalHasher.HashPatient() != modifiedHasher.HashPatient())
-      {
-        toStore.AddMetadata(ResourceType_Patient, metadataType, originalHasher.HashPatient());
-      }
-
-      assert(*it == originalHasher.HashInstance());
-      toStore.AddMetadata(ResourceType_Instance, metadataType, *it);
-
-
-      /**
-       * Store the resulting DICOM instance into the Orthanc store.
-       **/
-
-      std::string modifiedInstance;
-      if (context.Store(modifiedInstance, toStore) != StoreStatus_Success)
-      {
-        LOG(ERROR) << "Error while storing a modified instance " << *it;
-        throw OrthancException(ErrorCode_CannotStoreInstance);
-      }
-
-      // Sanity checks in debug mode
-      assert(modifiedInstance == modifiedHasher.HashInstance());
-
-
-      /**
-       * Compute the JSON object that is returned by the REST call.
-       **/
-
-      if (isFirst)
-      {
-        std::string newId;
-
-        switch (resourceType)
-        {
-          case ResourceType_Series:
-            newId = modifiedHasher.HashSeries();
-            break;
-
-          case ResourceType_Study:
-            newId = modifiedHasher.HashStudy();
-            break;
-
-          case ResourceType_Patient:
-            newId = modifiedHasher.HashPatient();
-            break;
-
-          default:
-            throw OrthancException(ErrorCode_InternalError);
-        }
-
-        result["Type"] = EnumerationToString(resourceType);
-        result["ID"] = newId;
-        result["Path"] = GetBasePath(resourceType, newId);
-        result["PatientID"] = modifiedHasher.HashPatient();
-        isFirst = false;
-      }
-    }
-
-    call.GetOutput().AnswerJson(result);
-  }
-
-
-
   static void ModifyInstance(RestApiPostCall& call)
   {
     DicomModification modification;
     modification.SetAllowManualIdentifiers(true);
 
-    if (ParseModifyRequest(modification, call))
-    {
-      if (modification.IsReplaced(DICOM_TAG_PATIENT_ID))
-      {
-        modification.SetLevel(ResourceType_Patient);
-      }
-      else if (modification.IsReplaced(DICOM_TAG_STUDY_INSTANCE_UID))
-      {
-        modification.SetLevel(ResourceType_Study);
-      }
-      else if (modification.IsReplaced(DICOM_TAG_SERIES_INSTANCE_UID))
-      {
-        modification.SetLevel(ResourceType_Series);
-      }
-      else
-      {
-        modification.SetLevel(ResourceType_Instance);
-      }
+    Json::Value request;
+    ParseModifyRequest(request, modification, call);
 
-      AnonymizeOrModifyInstance(modification, call);
+    if (modification.IsReplaced(DICOM_TAG_PATIENT_ID))
+    {
+      modification.SetLevel(ResourceType_Patient);
     }
+    else if (modification.IsReplaced(DICOM_TAG_STUDY_INSTANCE_UID))
+    {
+      modification.SetLevel(ResourceType_Study);
+    }
+    else if (modification.IsReplaced(DICOM_TAG_SERIES_INSTANCE_UID))
+    {
+      modification.SetLevel(ResourceType_Series);
+    }
+    else
+    {
+      modification.SetLevel(ResourceType_Instance);
+    }
+
+    AnonymizeOrModifyInstance(modification, call);
   }
 
 
@@ -471,48 +151,68 @@ namespace Orthanc
     DicomModification modification;
     modification.SetAllowManualIdentifiers(true);
 
-    if (ParseAnonymizationRequest(modification, call))
-    {
-      AnonymizeOrModifyInstance(modification, call);
-    }
+    Json::Value request;
+    ParseAnonymizationRequest(request, modification, call);
+
+    AnonymizeOrModifyInstance(modification, call);
   }
 
 
-  template <enum ChangeType changeType,
-            enum ResourceType resourceType>
+  static void SubmitModificationJob(std::auto_ptr<DicomModification>& modification,
+                                    bool isAnonymization,
+                                    RestApiPostCall& call,
+                                    const Json::Value& body,
+                                    ResourceType level)
+  {
+    ServerContext& context = OrthancRestApi::GetContext(call);
+
+    std::auto_ptr<ResourceModificationJob> job(new ResourceModificationJob(context));
+    
+    job->SetModification(modification.release(), level, isAnonymization);
+    job->SetOrigin(call);
+    
+    context.AddChildInstances(*job, call.GetUriComponent("id", ""));
+
+    OrthancRestApi::GetApi(call).SubmitCommandsJob
+      (call, job.release(), true /* synchronous by default */, body);
+  }
+
+
+  template <enum ResourceType resourceType>
   static void ModifyResource(RestApiPostCall& call)
   {
-    DicomModification modification;
+    std::auto_ptr<DicomModification> modification(new DicomModification);
 
-    if (ParseModifyRequest(modification, call))
-    {
-      modification.SetLevel(resourceType);
-      AnonymizeOrModifyResource(modification, MetadataType_ModifiedFrom, 
-                                changeType, resourceType, call);
-    }
+    Json::Value body;
+    ParseModifyRequest(body, *modification, call);
+
+    modification->SetLevel(resourceType);
+
+    SubmitModificationJob(modification, false /* not an anonymization */,
+                          call, body, resourceType);
   }
 
 
-  template <enum ChangeType changeType,
-            enum ResourceType resourceType>
+  template <enum ResourceType resourceType>
   static void AnonymizeResource(RestApiPostCall& call)
   {
-    DicomModification modification;
+    std::auto_ptr<DicomModification> modification(new DicomModification);
 
-    if (ParseAnonymizationRequest(modification, call))
-    {
-      AnonymizeOrModifyResource(modification, MetadataType_AnonymizedFrom, 
-                                changeType, resourceType, call);
-    }
+    Json::Value body;
+    ParseAnonymizationRequest(body, *modification, call);
+
+    SubmitModificationJob(modification, true /* anonymization */,
+                          call, body, resourceType);
   }
 
 
   static void StoreCreatedInstance(std::string& id /* out */,
                                    RestApiPostCall& call,
-                                   ParsedDicomFile& dicom)
+                                   ParsedDicomFile& dicom,
+                                   bool sendAnswer)
   {
     DicomInstanceToStore toStore;
-    toStore.SetRestOrigin(call);
+    toStore.SetOrigin(DicomInstanceOrigin::FromRest(call));
     toStore.SetParsedDicomFile(dicom);
 
     ServerContext& context = OrthancRestApi::GetContext(call);
@@ -522,6 +222,11 @@ namespace Orthanc
     {
       throw OrthancException(ErrorCode_CannotStoreInstance);
     }
+
+    if (sendAnswer)
+    {
+      OrthancRestApi::GetApi(call).AnswerStoredInstance(call, toStore, status);
+    }
   }
 
 
@@ -566,7 +271,7 @@ namespace Orthanc
   {
     if (tags.type() != Json::objectValue)
     {
-      throw OrthancException(ErrorCode_BadRequest);
+      throw OrthancException(ErrorCode_BadRequest, "Tags field is not an array");
     }
 
     // Inject the user-specified tags
@@ -591,7 +296,7 @@ namespace Orthanc
             tag != DICOM_TAG_STUDY_TIME &&
             dicom.HasTag(tag))
         {
-          throw OrthancException(ErrorCode_CreateDicomOverrideTag);
+          throw OrthancException(ErrorCode_CreateDicomOverrideTag, name);
         }
 
         if (tag == DICOM_TAG_PIXEL_DATA)
@@ -625,7 +330,7 @@ namespace Orthanc
     {
       for (Json::ArrayIndex i = 0; i < content.size(); i++)
       {
-        std::auto_ptr<ParsedDicomFile> dicom(base.Clone());
+        std::auto_ptr<ParsedDicomFile> dicom(base.Clone(false));
         const Json::Value* payload = NULL;
 
         if (content[i].type() == Json::stringValue)
@@ -657,7 +362,7 @@ namespace Orthanc
         dicom->ReplacePlainString(DICOM_TAG_INSTANCE_NUMBER, boost::lexical_cast<std::string>(i + 1));
         dicom->ReplacePlainString(DICOM_TAG_IMAGE_INDEX, boost::lexical_cast<std::string>(i + 1));
 
-        StoreCreatedInstance(someInstance, call, *dicom);
+        StoreCreatedInstance(someInstance, call, *dicom, false);
       }
     }
     catch (OrthancException&)
@@ -704,8 +409,8 @@ namespace Orthanc
         const char* tmp = request["Tags"]["SpecificCharacterSet"].asCString();
         if (!GetDicomEncoding(encoding, tmp))
         {
-          LOG(ERROR) << "Unknown specific character set: " << std::string(tmp);
-          throw OrthancException(ErrorCode_ParameterOutOfRange);
+          throw OrthancException(ErrorCode_ParameterOutOfRange,
+                                 "Unknown specific character set: " + std::string(tmp));
         }
       }
       else
@@ -735,6 +440,7 @@ namespace Orthanc
       // Select one existing child instance of the parent resource, to
       // retrieve all its tags
       Json::Value siblingTags;
+      std::string siblingInstanceId;
 
       {
         // Retrieve all the instances of the parent resource
@@ -747,7 +453,8 @@ namespace Orthanc
           throw OrthancException(ErrorCode_InternalError);
         }
 
-        context.ReadDicomAsJson(siblingTags, siblingInstances.front());
+        siblingInstanceId = siblingInstances.front();
+        context.ReadDicomAsJson(siblingTags, siblingInstanceId);
       }
 
 
@@ -758,11 +465,14 @@ namespace Orthanc
         if (siblingTags.isMember(SPECIFIC_CHARACTER_SET))
         {
           Encoding encoding;
+
           if (!siblingTags[SPECIFIC_CHARACTER_SET].isMember("Value") ||
               siblingTags[SPECIFIC_CHARACTER_SET]["Value"].type() != Json::stringValue ||
               !GetDicomEncoding(encoding, siblingTags[SPECIFIC_CHARACTER_SET]["Value"].asCString()))
           {
-            throw OrthancException(ErrorCode_CreateDicomParentEncoding);
+            LOG(WARNING) << "Instance with an incorrect Specific Character Set, "
+                         << "using the default Orthanc encoding: " << siblingInstanceId;
+            encoding = GetDefaultDicomEncoding();
           }
 
           dicom.SetEncoding(encoding);
@@ -809,7 +519,10 @@ namespace Orthanc
           else if (tag["Type"] == "String")
           {
             std::string value = tag["Value"].asString();
-            dicom.ReplacePlainString(*it, Toolbox::ConvertFromUtf8(value, dicom.GetEncoding()));
+
+            bool hasCodeExtensions;
+            Encoding encoding = dicom.DetectEncoding(hasCodeExtensions);
+            dicom.ReplacePlainString(*it, Toolbox::ConvertFromUtf8(value, encoding));
           }
         }
       }
@@ -831,7 +544,7 @@ namespace Orthanc
     
     // Inject time-related information
     std::string date, time;
-    SystemToolbox::GetNowDicom(date, time);
+    SystemToolbox::GetNowDicom(date, time, true /* use UTC time (not local time) */);
     dicom.ReplacePlainString(DICOM_TAG_ACQUISITION_DATE, date);
     dicom.ReplacePlainString(DICOM_TAG_ACQUISITION_TIME, time);
     dicom.ReplacePlainString(DICOM_TAG_CONTENT_DATE, date);
@@ -884,10 +597,7 @@ namespace Orthanc
     }
 
     std::string id;
-    StoreCreatedInstance(id, call, dicom);
-    OrthancRestApi::GetApi(call).AnswerStoredResource(call, id, ResourceType_Instance, StoreStatus_Success);
-
-    return;
+    StoreCreatedInstance(id, call, dicom, true);
   }
 
 
@@ -911,24 +621,146 @@ namespace Orthanc
       CreateDicomV1(dicom, call, request);
 
       std::string id;
-      StoreCreatedInstance(id, call, dicom);
-      OrthancRestApi::GetApi(call).AnswerStoredResource(call, id, ResourceType_Instance, StoreStatus_Success);
+      StoreCreatedInstance(id, call, dicom, true);
     }
   }
 
 
+  static void SplitStudy(RestApiPostCall& call)
+  {
+    ServerContext& context = OrthancRestApi::GetContext(call);
+
+    Json::Value request;
+    if (!call.ParseJsonRequest(request))
+    {
+      // Bad JSON request
+      throw OrthancException(ErrorCode_BadFileFormat);
+    }
+
+    const std::string study = call.GetUriComponent("id", "");
+
+    std::auto_ptr<SplitStudyJob> job(new SplitStudyJob(context, study));    
+    job->SetOrigin(call);
+
+    std::vector<std::string> series;
+    SerializationToolbox::ReadArrayOfStrings(series, request, "Series");
+
+    for (size_t i = 0; i < series.size(); i++)
+    {
+      job->AddSourceSeries(series[i]);
+    }
+
+    job->AddTrailingStep();
+
+    static const char* KEEP_SOURCE = "KeepSource";
+    if (request.isMember(KEEP_SOURCE))
+    {
+      job->SetKeepSource(SerializationToolbox::ReadBoolean(request, KEEP_SOURCE));
+    }
+
+    static const char* REMOVE = "Remove";
+    if (request.isMember(REMOVE))
+    {
+      if (request[REMOVE].type() != Json::arrayValue)
+      {
+        throw OrthancException(ErrorCode_BadFileFormat);
+      }
+
+      for (Json::Value::ArrayIndex i = 0; i < request[REMOVE].size(); i++)
+      {
+        if (request[REMOVE][i].type() != Json::stringValue)
+        {
+          throw OrthancException(ErrorCode_BadFileFormat);
+        }
+        else
+        {
+          job->Remove(FromDcmtkBridge::ParseTag(request[REMOVE][i].asCString()));
+        }
+      }
+    }
+
+    static const char* REPLACE = "Replace";
+    if (request.isMember(REPLACE))
+    {
+      if (request[REPLACE].type() != Json::objectValue)
+      {
+        throw OrthancException(ErrorCode_BadFileFormat);
+      }
+
+      Json::Value::Members tags = request[REPLACE].getMemberNames();
+
+      for (size_t i = 0; i < tags.size(); i++)
+      {
+        const Json::Value& value = request[REPLACE][tags[i]];
+        
+        if (value.type() != Json::stringValue)
+        {
+          throw OrthancException(ErrorCode_BadFileFormat);
+        }
+        else
+        {
+          job->Replace(FromDcmtkBridge::ParseTag(tags[i]), value.asString());
+        }
+      }
+    }
+
+    OrthancRestApi::GetApi(call).SubmitCommandsJob
+      (call, job.release(), true /* synchronous by default */, request);
+  }
+
+
+  static void MergeStudy(RestApiPostCall& call)
+  {
+    ServerContext& context = OrthancRestApi::GetContext(call);
+
+    Json::Value request;
+    if (!call.ParseJsonRequest(request))
+    {
+      // Bad JSON request
+      throw OrthancException(ErrorCode_BadFileFormat);
+    }
+
+    const std::string study = call.GetUriComponent("id", "");
+
+    std::auto_ptr<MergeStudyJob> job(new MergeStudyJob(context, study));    
+    job->SetOrigin(call);
+
+    std::vector<std::string> resources;
+    SerializationToolbox::ReadArrayOfStrings(resources, request, "Resources");
+
+    for (size_t i = 0; i < resources.size(); i++)
+    {
+      job->AddSource(resources[i]);
+    }
+
+    job->AddTrailingStep();
+
+    static const char* KEEP_SOURCE = "KeepSource";
+    if (request.isMember(KEEP_SOURCE))
+    {
+      job->SetKeepSource(SerializationToolbox::ReadBoolean(request, KEEP_SOURCE));
+    }
+
+    OrthancRestApi::GetApi(call).SubmitCommandsJob
+      (call, job.release(), true /* synchronous by default */, request);
+  }
+  
+
   void OrthancRestApi::RegisterAnonymizeModify()
   {
     Register("/instances/{id}/modify", ModifyInstance);
-    Register("/series/{id}/modify", ModifyResource<ChangeType_ModifiedSeries, ResourceType_Series>);
-    Register("/studies/{id}/modify", ModifyResource<ChangeType_ModifiedStudy, ResourceType_Study>);
-    Register("/patients/{id}/modify", ModifyResource<ChangeType_ModifiedPatient, ResourceType_Patient>);
+    Register("/series/{id}/modify", ModifyResource<ResourceType_Series>);
+    Register("/studies/{id}/modify", ModifyResource<ResourceType_Study>);
+    Register("/patients/{id}/modify", ModifyResource<ResourceType_Patient>);
 
     Register("/instances/{id}/anonymize", AnonymizeInstance);
-    Register("/series/{id}/anonymize", AnonymizeResource<ChangeType_AnonymizedSeries, ResourceType_Series>);
-    Register("/studies/{id}/anonymize", AnonymizeResource<ChangeType_AnonymizedStudy, ResourceType_Study>);
-    Register("/patients/{id}/anonymize", AnonymizeResource<ChangeType_AnonymizedPatient, ResourceType_Patient>);
+    Register("/series/{id}/anonymize", AnonymizeResource<ResourceType_Series>);
+    Register("/studies/{id}/anonymize", AnonymizeResource<ResourceType_Study>);
+    Register("/patients/{id}/anonymize", AnonymizeResource<ResourceType_Patient>);
 
     Register("/tools/create-dicom", CreateDicom);
+
+    Register("/studies/{id}/split", SplitStudy);
+    Register("/studies/{id}/merge", MergeStudy);
   }
 }
