@@ -2,7 +2,7 @@
  * Orthanc - A Lightweight, RESTful DICOM Store
  * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
  * Department, University Hospital of Liege, Belgium
- * Copyright (C) 2017-2018 Osimis S.A., Belgium
+ * Copyright (C) 2017-2019 Osimis S.A., Belgium
  *
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -34,10 +34,13 @@
 #include "../PrecompiledHeadersServer.h"
 #include "ArchiveJob.h"
 
+#include "../../Core/Cache/SharedArchive.h"
 #include "../../Core/Compression/HierarchicalZipWriter.h"
 #include "../../Core/DicomParsing/DicomDirWriter.h"
 #include "../../Core/Logging.h"
 #include "../../Core/OrthancException.h"
+#include "../OrthancConfiguration.h"
+#include "../ServerContext.h"
 
 #include <stdio.h>
 
@@ -47,7 +50,12 @@
 
 static const uint64_t MEGA_BYTES = 1024 * 1024;
 static const uint64_t GIGA_BYTES = 1024 * 1024 * 1024;
-static const char* MEDIA_IMAGES_FOLDER = "IMAGES"; 
+
+static const char* const MEDIA_IMAGES_FOLDER = "IMAGES"; 
+static const char* const KEY_DESCRIPTION = "Description";
+static const char* const KEY_INSTANCES_COUNT = "InstancesCount";
+static const char* const KEY_UNCOMPRESSED_SIZE_MB = "UncompressedSizeMB";
+
 
 namespace Orthanc
 {
@@ -778,11 +786,9 @@ namespace Orthanc
   };
 
 
-  ArchiveJob::ArchiveJob(boost::shared_ptr<TemporaryFile>& target,
-                         ServerContext& context,
+  ArchiveJob::ArchiveJob(ServerContext& context,
                          bool isMedia,
                          bool enableExtendedSopClass) :
-    target_(target),
     context_(context),
     archive_(new ArchiveIndex(ResourceType_Patient)),  // root
     isMedia_(isMedia),
@@ -791,60 +797,159 @@ namespace Orthanc
     instancesCount_(0),
     uncompressedSize_(0)
   {
-    if (target.get() == NULL)
+  }
+
+  
+  ArchiveJob::~ArchiveJob()
+  {
+    if (!mediaArchiveId_.empty())
     {
-      throw OrthancException(ErrorCode_NullPointer);
+      context_.GetMediaArchive().Remove(mediaArchiveId_);
     }
   }
 
 
+  void ArchiveJob::SetSynchronousTarget(boost::shared_ptr<TemporaryFile>& target)
+  {
+    if (target.get() == NULL)
+    {
+      throw OrthancException(ErrorCode_NullPointer);
+    }
+    else if (writer_.get() != NULL ||  // Already started
+             synchronousTarget_.get() != NULL ||
+             asynchronousTarget_.get() != NULL)
+    {
+      throw OrthancException(ErrorCode_BadSequenceOfCalls);
+    }
+    else
+    {
+      synchronousTarget_ = target;
+    }
+  }
+
+
+  void ArchiveJob::SetDescription(const std::string& description)
+  {
+    if (writer_.get() != NULL)   // Already started
+    {
+      throw OrthancException(ErrorCode_BadSequenceOfCalls);
+    }
+    else
+    {
+      description_ = description;
+    }
+  }
+
+  
   void ArchiveJob::AddResource(const std::string& publicId)
   {
     if (writer_.get() != NULL)   // Already started
     {
       throw OrthancException(ErrorCode_BadSequenceOfCalls);
     }
-        
-    ResourceIdentifiers resource(context_.GetIndex(), publicId);
-    archive_->Add(context_.GetIndex(), resource);
+    else
+    {
+      ResourceIdentifiers resource(context_.GetIndex(), publicId);
+      archive_->Add(context_.GetIndex(), resource);
+    }
   }
 
   
-  void ArchiveJob::SignalResubmit()
+  void ArchiveJob::Reset()
   {
-    LOG(ERROR) << "Cannot resubmit the creation of an archive";
-    throw OrthancException(ErrorCode_BadSequenceOfCalls);
+    throw OrthancException(ErrorCode_BadSequenceOfCalls,
+                           "Cannot resubmit the creation of an archive");
   }
 
   
   void ArchiveJob::Start()
   {
+    TemporaryFile* target = NULL;
+    
+    if (synchronousTarget_.get() == NULL)
+    {
+      {
+        OrthancConfiguration::ReaderLock lock;
+        asynchronousTarget_.reset(lock.GetConfiguration().CreateTemporaryFile());
+      }
+
+      target = asynchronousTarget_.get();
+    }
+    else
+    {
+      target = synchronousTarget_.get();
+    }
+
+    assert(target != NULL);
+    target->Touch();  // Make sure we can write to the temporary file
+    
     if (writer_.get() != NULL)
     {
       throw OrthancException(ErrorCode_BadSequenceOfCalls);
     }
 
-    writer_.reset(new ZipWriterIterator(*target_, context_, *archive_,
+    writer_.reset(new ZipWriterIterator(*target, context_, *archive_,
                                         isMedia_, enableExtendedSopClass_));
 
     instancesCount_ = writer_->GetInstancesCount();
     uncompressedSize_ = writer_->GetUncompressedSize();
   }
 
+
+
+  namespace
+  {
+    class DynamicTemporaryFile : public IDynamicObject
+    {
+    private:
+      std::auto_ptr<TemporaryFile>   file_;
+
+    public:
+      DynamicTemporaryFile(TemporaryFile* f) : file_(f)
+      {
+        if (f == NULL)
+        {
+          throw OrthancException(ErrorCode_NullPointer);
+        }
+      }
+
+      const TemporaryFile& GetFile() const
+      {
+        assert(file_.get() != NULL);
+        return *file_;
+      }
+    };
+  }
   
-  JobStepResult ArchiveJob::ExecuteStep()
+
+  void ArchiveJob::FinalizeTarget()
+  {
+    writer_.reset();  // Flush all the results
+
+    if (asynchronousTarget_.get() != NULL)
+    {
+      // Asynchronous behavior: Move the resulting file into the media archive
+      mediaArchiveId_ = context_.GetMediaArchive().Add(
+        new DynamicTemporaryFile(asynchronousTarget_.release()));
+    }
+  }
+    
+
+  JobStepResult ArchiveJob::Step()
   {
     assert(writer_.get() != NULL);
 
-    if (target_.unique())
+    if (synchronousTarget_.get() != NULL &&
+        synchronousTarget_.unique())
     {
       LOG(WARNING) << "A client has disconnected while creating an archive";
-      return JobStepResult::Failure(ErrorCode_NetworkProtocol);          
+      return JobStepResult::Failure(ErrorCode_NetworkProtocol,
+                                    "A client has disconnected while creating an archive");
     }
         
     if (writer_->GetStepsCount() == 0)
     {
-      writer_.reset();  // Flush all the results
+      FinalizeTarget();
       return JobStepResult::Success();
     }
     else
@@ -855,7 +960,7 @@ namespace Orthanc
 
       if (currentStep_ == writer_->GetStepsCount())
       {
-        writer_.reset();  // Flush all the results
+        FinalizeTarget();
         return JobStepResult::Success();
       }
       else
@@ -893,12 +998,41 @@ namespace Orthanc
     }
   }
 
-    
+
   void ArchiveJob::GetPublicContent(Json::Value& value)
   {
-    value["Description"] = description_;
-    value["InstancesCount"] = instancesCount_;
-    value["UncompressedSizeMB"] =
+    value = Json::objectValue;
+    value[KEY_DESCRIPTION] = description_;
+    value[KEY_INSTANCES_COUNT] = instancesCount_;
+    value[KEY_UNCOMPRESSED_SIZE_MB] =
       static_cast<unsigned int>(uncompressedSize_ / MEGA_BYTES);
   }
+
+
+  bool ArchiveJob::GetOutput(std::string& output,
+                             MimeType& mime,
+                             const std::string& key)
+  {   
+    if (key == "archive" &&
+        !mediaArchiveId_.empty())
+    {
+      SharedArchive::Accessor accessor(context_.GetMediaArchive(), mediaArchiveId_);
+
+      if (accessor.IsValid())
+      {
+        const DynamicTemporaryFile& f = dynamic_cast<DynamicTemporaryFile&>(accessor.GetItem());
+        f.GetFile().Read(output);
+        mime = MimeType_Zip;
+        return true;
+      }
+      else
+      {
+        return false;
+      }
+    }    
+    else
+    {
+      return false;
+    }
+  }
 }
